/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"

#include <librnd/core/actions.h>
#include <librnd/core/rotate.h>
#include <librnd/hid/hid.h>

#include "conf_core.h"
#include "data.h"
#include "search.h"
#include "obj_any.h"

#include "buffer.h"

osm_layer_t osm_buffer[OSM_BUFFER_MAX];

void osm_buffer_clear(osm_layer_t *buff)
{
	osm_layer_clear_remove_list(buff);
	osm_layer_clear_objs(buff);
}


void osm_buffer_copy_selected(osm_layer_t *src, osm_layer_t *buff, rnd_coord_t dx, rnd_coord_t dy)
{
	vtp0_t tmp = {0}, tmp2 = {0};
	long n, m;
	int t;

	osm_buffer_clear(buff);

	osm_search_all_selected(&tmp, src);

	/* pick up unselected referees: can't copy a way without its ndoes */
	for(n = 0; n < tmp.used; n++) {
		osm_any_obj_t *oobj = tmp.array[n];

		tmp2.used = 0;
		osm_obj_referee_objs(oobj, &tmp2, 0);

		for(m = 0; m < tmp2.used; m++) {
			osm_any_obj_t *ref = tmp2.array[m];
			if (!ref->hdr.selected) {
				vtp0_append(&tmp, ref);
				ref->hdr.selected = 1;
			}
		}
	}

	/* copy object to buffer going by object types */
	for(t = 0; t < OSM_OBJ_max; t++) {
		for(n = 0; n < tmp.used; n++) {
			osm_any_obj_t *nobj, *oobj = tmp.array[n];

			if (oobj->hdr.type != t) continue; /* copy nodes first then ways then relations so that references are always valid */

			oobj->hdr.selected = oobj->hdr.referer = 0; /* clean up temporary used flags */
			nobj = osm_obj_dup_into(oobj, buff, 0);
			if ((t == OSM_OBJ_NODE) && ((dx != 0) || (dy != 0))) {
				nobj->node.att_x = nobj->node.x - dx;
				nobj->node.att_y = nobj->node.y - dy;
				osm_node_move_to_att(&nobj->node, 0);
			}
		}
	}

	vtp0_uninit(&tmp);
	vtp0_uninit(&tmp2);
}

void osm_buffer_paste(osm_layer_t *dst, osm_layer_t *buff, rnd_coord_t dx, rnd_coord_t dy)
{
	htdr_t dr = {0};
	htIp_entry_t *e;
	int t;

	htdr_init(&dr, htdr_keyhash, htdr_keyeq);

	/* calculate id renumbering into dr */
	for(t = 0; t < OSM_OBJ_max; t++) {
		for(e = htIp_first(&buff->id2obj[t]); e != NULL; e = htIp_next(&buff->id2obj[t], e)) {
			osm_any_obj_t *o = e->value;
			htdr_key_t drk;
			htdr_value_t drv;

			drk.type = o->hdr.type;
			drk.old_id = o->hdr.id;
			drv.valid = 1;
			drv.new_id = --(dst->last_new_id);
			drv.new_version = 1;
			htdr_set(&dr, drk, drv);
		}
	}

	for(t = 0; t < OSM_OBJ_max; t++) {
		for(e = htIp_first(&buff->id2obj[t]); e != NULL; e = htIp_next(&buff->id2obj[t], e)) {
			osm_any_obj_t *nobj, *oobj = e->value;
			htdr_key_t drk;
			htdr_value_t drv;

			/* look up the new ID */
			drk.type = oobj->hdr.type;
			drk.old_id = oobj->hdr.id;
			drv = htdr_get(&dr, drk);

			nobj = osm_obj_dup_into(oobj, dst, drv.new_id);
			if ((nobj->hdr.type == OSM_OBJ_NODE) && ((dx != 0) || (dy != 0))) {
				nobj->node.att_x = nobj->node.x + dx;
				nobj->node.att_y = nobj->node.y + dy;
				osm_node_move_to_att(&nobj->node, 0);
			}

			nobj->hdr.calls->diffresult_referees(nobj, &dr, 0, 1);
		}
	}

	htdr_uninit(&dr);
}

void osm_buffer_xor_draw(osm_layer_t *dst, osm_layer_t *buff, rnd_coord_t dx, rnd_coord_t dy, rnd_hid_gc_t gc)
{
	htIp_entry_t *e;
	int t;

	/* move all nodes' attached coord to where we render */
	for(e = htIp_first(&buff->id2obj[OSM_OBJ_NODE]); e != NULL; e = htIp_next(&buff->id2obj[OSM_OBJ_NODE], e)) {
		osm_node_t *node = e->value;
		node->att_x = node->x + dx;
		node->att_y = node->y + dy;
	}

	for(t = 0; t < OSM_OBJ_max; t++)
		for(e = htIp_first(&buff->id2obj[t]); e != NULL; e = htIp_next(&buff->id2obj[t], e))
			osm_obj_draw_attached(e->value, gc);
}


static const char osm_acts_BufferCopy[] = "BufferCopy([x, y])";
static const char osm_acth_BufferCopy[] = "Copy selection to currently active paste buffer.\n";
fgw_error_t osm_act_BufferCopy(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	osm_design_t *osm = (osm_design_t *)RND_ACT_DESIGN;
	osm_layer_t *ly = osm_layer_get_curr(osm);
	rnd_coord_t x = osm->crosshair_x, y = osm->crosshair_y;

	RND_ACT_MAY_CONVARG(1, FGW_COORD, BufferCopy, x = fgw_coord(&(argv[1])));
	RND_ACT_MAY_CONVARG(2, FGW_COORD, BufferCopy, y = fgw_coord(&(argv[2])));

	osm_buffer_copy_selected(ly, OSM_PASTEBUFFER, x, y);

	RND_ACT_IRES(0);
	return 0;
}

static const char osm_acts_BufferRotate90[] = "BufferRotate90([n])";
static const char osm_acth_BufferRotate90[] = "Rotate the buffer n times 90 degrees\n";
fgw_error_t osm_act_BufferRotate90(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	osm_layer_t *buff = OSM_PASTEBUFFER;
	htIp_entry_t *e;
	int n = 1;

	RND_ACT_MAY_CONVARG(1, FGW_COORD, BufferRotate90, n = argv[1].val.nat_int);

	n = n % 4;

	for(e = htIp_first(&buff->id2obj[OSM_OBJ_NODE]); e != NULL; e = htIp_next(&buff->id2obj[OSM_OBJ_NODE], e)) {
		osm_node_t *node = e->value;
		node->att_x = node->x;
		node->att_y = node->y;
		RND_COORD_ROTATE90(node->att_x, node->att_y,   0,0,  n);
		osm_node_move_to_att(node, 0);
	}

	return 0;
}

static const char osm_acts_BufferChk[] = "BufferChk(num)";
static const char osm_acth_BufferChk[] = "Returns whether buffer num is active or not\n";
fgw_error_t osm_act_BufferChk(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	int expected;
	RND_ACT_CONVARG(1, FGW_INT, BufferChk, expected = argv[1].val.nat_int);
	RND_ACT_IRES((conf_core.editor.buffer_number + 1) == expected);
	return 0;
}

static const char osm_acts_BufferSwitch[] = "BufferSwitch(num)";
static const char osm_acth_BufferSwitch[] = "Activates buffer num\n";
fgw_error_t osm_act_BufferSwitch(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	int bn;
	RND_ACT_CONVARG(1, FGW_INT, BufferSwitch, bn = argv[1].val.nat_int);

	rnd_conf_set_design("editor/buffer_number", "%d", bn-1);
	rnd_gui->invalidate_all(rnd_gui);

	RND_ACT_IRES(0);
	return 0;
}

static rnd_action_t buffer_action_list[] = {
/*
	{"BufferPaste", osm_act_BufferPaste, osm_acth_BufferPaste, osm_acts_BufferPaste},*/
	{"BufferCopy", osm_act_BufferCopy, osm_acth_BufferCopy, osm_acts_BufferCopy},
/*	{"BufferCut", osm_act_BufferCut, osm_acth_BufferCut, osm_acts_BufferCut},
	{"BufferClear", osm_act_BufferClear, osm_acth_BufferClear, osm_acts_BufferClear},
	{"BufferSave", osm_act_BufferSave, osm_acth_BufferSave, osm_acts_BufferSave},
	{"BufferLoad", osm_act_BufferLoad, osm_acth_BufferLoad, osm_acts_BufferLoad},
*/
	{"BufferRotate90", osm_act_BufferRotate90, osm_acth_BufferRotate90, osm_acts_BufferRotate90},
	{"BufferChk", osm_act_BufferChk, osm_acth_BufferChk, osm_acts_BufferChk},
	{"BufferSwitch", osm_act_BufferSwitch, osm_acth_BufferSwitch, osm_acts_BufferSwitch}
};


void osm_buffer_init2(void)
{
	int n;
	for(n = 0; n < OSM_BUFFER_MAX; n++)
		osm_layer_init(&osm_buffer[n]);
	RND_REGISTER_ACTIONS(buffer_action_list, NULL);
}

void osm_buffer_uninit(void)
{
	int n;
	for(n = 0; n < OSM_BUFFER_MAX; n++)
		osm_layer_free_fields(&osm_buffer[n]);
}



