/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include <librnd/config.h>

#include <assert.h>

#include <librnd/core/event.h>
#include <librnd/core/compat_misc.h>

#include "data.h"
#include "obj_any.h"
#include <librnd/core/box.h>
#include "conf_core.h"
#include "event.h"

osm_design_t osm;

char osm_measurement_layer_name[] = "measure";


void osm_layer_init(osm_layer_t *layer)
{
	int n;
	memset(layer, 0, sizeof(osm_layer_t));
	osm_rtree_init(&layer->objs);
	for(n = 0; n < OSM_ZORD_MAX; n++)
		osm_rtree_init(&layer->zord[n]);

	layer->vis = 1;
	for(n = 0; n < OSM_OBJ_max; n++)
		htIp_init(&layer->id2obj[n], htIp_keyhash, htIp_keyeq);

	uundo_list_init(&layer->undo);
	vtp0_init(&layer->remove_list);
}

osm_layer_t *osm_layer_new(void)
{
	osm_layer_t *res = malloc(sizeof(osm_layer_t));
	osm_layer_init(res);
	return res;
}

long osm_layer_append_to_design(osm_design_t *osm, osm_layer_t *layer, int bottom)
{
	long res = -1;

	assert(layer->parent == NULL);
	res = osm->layers.used;
	if (bottom) {
		vtp0_insert_len(&osm->layers, 0, (void **)&layer, 1);
		osm->lysel++;
	}
	else
		vtp0_append(&osm->layers, layer);
	layer->parent = osm;
	return res;
}

static osm_layer_t *osm_layer_by_name_(osm_design_t *osm, const char *name, int alloc, int bottom)
{
	int n;
	osm_layer_t *ly;

	for(n = 0; n < osm->layers.used; n++) {
		ly = osm->layers.array[n];
		if ((ly->name != NULL) && (strcmp(ly->name, name) == 0))
			return osm->layers.array[n];
	}

	if (!alloc)
		return NULL;

	ly = osm_layer_new();
	osm_layer_append_to_design(osm, ly, bottom);
	ly->name = rnd_strdup(name);
	rnd_event(&osm->hidlib, OSM_EVENT_LAYERS_CHANGED, NULL);
	return ly;
}

osm_layer_t *osm_layer_by_name(osm_design_t *osm, const char *name, int alloc)
{
	return osm_layer_by_name_(osm, name, alloc, 0);
}

osm_layer_t *osm_layer_by_name_bottom(osm_design_t *osm, const char *name, int alloc)
{
	return osm_layer_by_name_(osm, name, alloc, 1);
}

void osm_layer_clear_remove_list(osm_layer_t *ly)
{
	long n;

	for(n = 0; n < ly->remove_list.used; n++)
		if (ly->remove_list.array[n] != NULL)
			osm_obj_free(ly->remove_list.array[n]);

	vtp0_uninit(&ly->remove_list);
}

void osm_layer_clear_objs(osm_layer_t *ly)
{
	void *o;
	long n;
	osm_rtree_it_t it;

	for(n = 0; n < OSM_ZORD_MAX; n++) {
		osm_rtree_uninit(&ly->zord[n]);
		osm_rtree_init(&ly->zord[n]);
	}

	for(o = osm_rtree_all_first(&it, &ly->objs); o != NULL; o = osm_rtree_all_next(&it))
		osm_obj_free(o);

	osm_rtree_uninit(&ly->objs);
	osm_rtree_init(&ly->objs);
	for(n = 0; n < OSM_OBJ_max; n++)
		htIp_clear(&ly->id2obj[n]);
}

void osm_layer_free_fields(osm_layer_t *ly)
{
	long n;
	void *o;
	osm_rtree_it_t it;


	uundo_list_uninit(&ly->undo);

	free(ly->name); ly->name = NULL;
	free(ly->short_name); ly->short_name = NULL;

	if (ly->parent != NULL) {
		osm_design_t *osm = ly->parent;
		/* optimization: go backward because the most common case is removing the last layer */
		for(n = osm->layers.used - 1; n >= 0; n--) {
			if (osm->layers.array[n] == ly) {
				vtp0_remove(&osm->layers, n, 1);
				break;
			}
		}
		ly->parent = NULL;
	}

	for(o = osm_rtree_all_first(&it, &ly->objs); o != NULL; o = osm_rtree_all_next(&it))
		osm_obj_free(o);
	osm_rtree_uninit(&ly->objs);

	for(n = 0; n < OSM_OBJ_max; n++)
		htIp_uninit(&ly->id2obj[n]);

	osm_layer_clear_remove_list(ly);
}

void osm_layer_destroy(osm_layer_t *ly)
{
	osm_layer_free_fields(ly);
	free(ly);
}


void osm_design_free_fields(osm_design_t *osm)
{
	long n;

	/* optimization: go backward to avoid memmoves */
	for(n = osm->layers.used - 1; n >= 0; n--)
		osm_layer_destroy(osm->layers.array[n]);
	vtp0_uninit(&osm->layers);
}

int osm_layer_set_vis(osm_design_t *osm, rnd_cardinal_t lid, int vis, int emit_event)
{
	if (lid >= osm->layers.used)
		return -1;

	((osm_layer_t *)osm->layers.array[lid])->vis = vis;

	if (emit_event)
		rnd_event(&osm->hidlib, OSM_EVENT_LAYERVIS_CHANGED, NULL);
	return 0;
}

int osm_is_empty(osm_design_t *osm)
{
	long lid;

	for(lid = 0; lid < osm->layers.used; lid++)
		if (((osm_layer_t *)osm->layers.array[lid])->objs.size != 0)
			return 0;
	return 1;
}

void osm_data_bbox(osm_design_t *osm)
{
	long lid;

	osm->bbox.x1 = osm->bbox.y1 = RND_COORD_MAX;
	osm->bbox.x2 = osm->bbox.y2 = -RND_COORD_MAX;
	for(lid = 0; lid < osm->layers.used; lid++) {
		osm_layer_t *ly = (osm_layer_t *)osm->layers.array[lid];
		if (ly->objs.size != 0) {
			rnd_box_bump_point((rnd_box_t *)&osm->bbox, ly->objs.bbox.x1, ly->objs.bbox.y1);
			rnd_box_bump_point((rnd_box_t *)&osm->bbox, ly->objs.bbox.x2, ly->objs.bbox.y2);
		}
	}

	if ((osm->bbox.x1 == RND_COORD_MAX) || (osm->bbox.y1 == RND_COORD_MAX) || (osm->bbox.x2 == -RND_COORD_MAX) || (osm->bbox.y2 == -RND_COORD_MAX)) {
		/* empty layer - make it 1*1 mm so it's easier to handle */
		osm->bbox.x1 = osm->bbox.y1 = 0;
		osm->bbox.x2 = osm->bbox.y2 = RND_MM_TO_COORD(1);
	}
}

void osm_layer_select(osm_design_t *osm, int idx)
{
	osm->lysel = idx;
	rnd_event(&osm->hidlib, OSM_EVENT_LAYER_SELECTED, "i", idx);
}

void osm_data_reverse_layers(osm_design_t *osm, int inhibit_ev)
{
	long n, m;

	for(n = 0, m = osm->layers.used-1; n < osm->layers.used/2; n++,m--) {
		osm_layer_t *ly = osm->layers.array[n];
		osm->layers.array[n] = osm->layers.array[m];
		osm->layers.array[m] = ly;
	}

	if (!inhibit_ev)
		rnd_event(&osm->hidlib, OSM_EVENT_LAYERS_CHANGED, NULL);
}

osm_id_t osm_str2id(const char *str, char **end)
{
	if (str == NULL) {
		*end = "NULL";
		return 0;
	}

	return osm_strtoid(str, end, 10);
}

osm_layer_t *osm_layer_get_curr(osm_design_t *osm)
{
	if ((osm->lysel >= 0) && (osm->lysel < osm->layers.used))
		return osm->layers.array[osm->lysel];
	return NULL;
}
