/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023,2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#ifndef OSM_OBJ_ANY_H
#define OSM_OBJ_ANY_H

#include <assert.h>
#include <time.h>

#include "config.h"
#include "osm_typedefs.h"
#include "data.h"
#include "tags.h"
#include "htdr.h"

#include <librnd/core/global_typedefs.h>
#include "rtree.h"

extern const char *osm_obj_name[]; /* indexed by osm_obtype_t*/


typedef struct osm_objcalls_s {
	osm_any_obj_t *(*alloc)(void);
	void (*free_fields)(osm_any_obj_t *obj);
	void (*draw)(osm_any_obj_t *obj, rnd_hid_gc_t gc, const rnd_hid_expose_ctx_t *region, rnd_xform_t *xform, int overlay);
	void (*draw_attached)(osm_any_obj_t *obj, rnd_hid_gc_t gc);
	void (*bbox)(osm_any_obj_t *obj); /* Return object bbox (calculates bbox if needed, doesn't change rtree) */
	int (*isc_box)(const osm_any_obj_t *o, const osm_rtree_box_t *box);
	osm_tags_t *(*get_tags)(osm_any_obj_t *o);
	void (*pre_edit)(osm_any_obj_t *o, int geo, int auto_fork);
	void (*post_edit)(osm_any_obj_t *o, int tags_changed);
	void (*referee_objs)(osm_any_obj_t *o, vtp0_t *dst, int update_att);
	long (*osm_way_referee_idx)(osm_any_obj_t *parent, osm_any_obj_t *child); /* return the index of child within the lists of parent or -1 if not present */
	void (*diffresult)(osm_any_obj_t *o, osm_id_t new_id, osm_id_t new_version);   /* change the id/version of an object while processing a diffresult */
	void (*diffresult_referees)(osm_any_obj_t *o, htdr_t *dr, int permit_del, int undoable); /* renumber all referee IDs of an object from dr */
	void (*pre_remove_referee)(osm_any_obj_t *o, osm_objtype_t rem_ty, osm_id_t rem_id, int undoable); /* called after the 'rem' object has been removed so that referers can remove references */
	void (*post_remove)(osm_any_obj_t *o, int undoable); /* called after an object have been removed; the object should clean up orphans (e.g. ways cleaning up unused nodes); called only if the object ended up on the layer's remove-list */
	osm_any_obj_t *(*dup_into)(osm_any_obj_t *o, osm_layer_t *dst, osm_id_t new_id);
	const char *(*visual_type)(osm_any_obj_t *o); /* return the visual type name of an object */
} osm_objcalls_t;


typedef struct osm_hdr_s {
	osm_rtree_box_t bbox;
	osm_objtype_t type;
	const osm_objcalls_t *calls;

	/* current state of the object within layer */
	char zord;               /* index of the zord rtree the object is first rendered in (bottom); should draw main object features; -1 if not in rtree */
	unsigned bbox_valid:1;   /* 1 if bbox is calculated and valid */
	unsigned dirty:1;        /* referee objects changed, need to recalculate caches and bbox */
	unsigned in_objs:1;      /* 1 if obj is added to layer's ->objs rtree */

	unsigned selected:1;
	unsigned referer:1;      /* used as a temporary mark while collecting referers to make sure each referer is found and added only once */
	void *tmp_grp[4];        /* temporary grouping; used by some plugins */

	osm_style_t *style; /* contains target zord */

	/* official data model */
	osm_id_t id;          /* object ID */
	osm_id_t uid;         /* user ID */
	time_t timestamp;
	unsigned visible:1;   /* 0 when deleted */
	long version;
	osm_id_t changeset;

	/* cache */
	osm_layer_t *layer;
} osm_hdr_t;

#include "obj_node.h"
#include "obj_way.h"
#include "obj_relation.h"
#include "obj_changeset.h"

union osm_any_obj_u {
	osm_hdr_t hdr;
	osm_node_t node;
	osm_way_t way;
	osm_relation_t relation;
	osm_changeset_t changeset;
};

/* High level: undoable object creation; after this, undoable modify calls
   can be used to fill in data of the new object. */
osm_any_obj_t *osm_obj_create(osm_layer_t *layer, osm_objtype_t type, int undoable);

/* High level: undoable object removal; removes object from its current layer */
void osm_obj_remove(osm_any_obj_t *obj, int undoable);

/* High level: undoable tag value modification */
void osm_obj_modify_tag(osm_layer_t *layer, osm_any_obj_t *obj, const char *key, const char *new_val, int undoable);

/* High level: create or remove tag */
void osm_tag_create(osm_any_obj_t *obj, const char *key, const char *new_val, int undoable);
void osm_tag_remove(osm_any_obj_t *obj, const char *key, int undoable);



/* Low level: add obj in layer's rtrees and hash and set obj->ly. */
void osm_obj_add_to_layer(osm_layer_t *ly, osm_any_obj_t *obj);

/* Low level: remove object from layer's rtrees and hash, sets obj->ly to NULL.
   Does not free obj. */
void osm_obj_del_from_layer(osm_any_obj_t *obj);


/* Append objects that macth refmask and are referring to obj to lst, also
   mark them with ->hdr.referer. Typical call sequence: buil a full list by
   callin osm_obj_list_referers() multiple times then call
   osm_obj_list_clear_referer() to finalize it. The list is valid until
   any change to the layer. */
long osm_obj_list_referers(vtp0_t *lst, osm_any_obj_t *obj, osm_objtype_mask_t resmask);

/* Similar to osm_obj_list_referers() but instead of listing, mark any referer
   dirty so the referer's cache and bbox gets recalculated */
long osm_obj_dirt_referers(osm_any_obj_t *obj, osm_objtype_mask_t mask);


/* Clear ->hdr.referer on each object in lst */
void osm_obj_list_clear_referer(vtp0_t *lst);

/* Similar to strtod(): converts object type name to object type */
osm_objtype_t osm_str_to_obj_type(const char *str, char **end);

/* Convert a type:#id pair; whitespace may follow type; both the colon and
   the hashmark are optional. Returns OSM_OBJ_invalid on error */
osm_objtype_t osm_str_to_obj_type_id(const char *str, char **end, osm_id_t *id_out);


/* Look up object by type:id within a layer; may return NULL */
RND_INLINE osm_any_obj_t *osm_obj_by_id(osm_layer_t *ly, osm_objtype_t obj_type, osm_id_t obj_id);

/* Find all referee nodes; plus:
    - if dst is not NULL and obj is not already marked as ->hdr.referer,
      mark obj and add to dst
    - if update_att is 1, copy x;y to att_x;att_y in any node referred by
      obj, in perparation of attached draw */
RND_INLINE void osm_obj_referee_objs(osm_any_obj_t *obj, vtp0_t *dst, int update_att);

RND_INLINE osm_layer_t *osm_obj_parent_layer(const osm_any_obj_t *obj);
RND_INLINE osm_design_t *osm_obj_parent_design(const osm_any_obj_t *obj);

/* Return object bbox (calculates bbox if needed, doesn't change rtree) */
RND_INLINE osm_rtree_box_t *osm_obj_bbox(osm_any_obj_t *obj);

/* Force-recalculate bbox and update rtrees */
osm_rtree_box_t *osm_obj_bbox_update(osm_any_obj_t *obj);

/* Call before and after editing the ->curr of an object. The pre call
   forks the object if it's still the original. Geo should be 0 if no
   geometry is going the be changed (typically: only tag changes). */
RND_INLINE void osm_obj_pre_edit(osm_any_obj_t *obj, int geo);
RND_INLINE void osm_obj_post_edit(osm_any_obj_t *obj, int tags_changed);

/* Same as osm_obj_pre_edit_*() but does not fork ->curr, assuming
   the caller is going to edit ->orig - useful only when re-importing
   data from a server */
RND_INLINE void osm_obj_pre_edit_orig(osm_any_obj_t *obj, int geo);
RND_INLINE void osm_obj_post_edit_orig(osm_any_obj_t *obj, int tags_changed);

RND_INLINE void osm_obj_draw_attached(osm_any_obj_t *obj, rnd_hid_gc_t gc);

/* Dup obj into dst, keeping the original ID and references (not copying
   referees). New object's id is new_id if it is not 0 (else the id is copied) */
RND_INLINE osm_any_obj_t *osm_obj_dup_into(osm_any_obj_t *obj, osm_layer_t *dst, osm_id_t new_id);

/*** implementation (inlines) ***/

RND_INLINE osm_layer_t *osm_obj_parent_layer(const osm_any_obj_t *obj)
{
	return obj->hdr.layer;
}

RND_INLINE osm_design_t *osm_obj_parent_design(const osm_any_obj_t *obj)
{
	osm_layer_t *ly = osm_obj_parent_layer(obj);
	if (ly == NULL)
		return NULL;
	return ly->parent;
}

RND_INLINE osm_rtree_box_t *osm_obj_bbox(osm_any_obj_t *obj)
{
	if (!obj->hdr.bbox_valid)
		obj->hdr.calls->bbox(obj);
	assert(obj->hdr.bbox_valid);
	return &obj->hdr.bbox;
}

RND_INLINE void osm_obj_free(osm_any_obj_t *obj)
{
	obj->hdr.calls->free_fields(obj);
	free(obj);
}

RND_INLINE void osm_obj_pre_edit(osm_any_obj_t *obj, int geo)
{
	obj->hdr.calls->pre_edit(obj, geo, 1);
}

RND_INLINE void osm_obj_post_edit(osm_any_obj_t *obj, int tags_changed)
{
	obj->hdr.calls->post_edit(obj, tags_changed);
}


RND_INLINE void osm_obj_pre_edit_orig(osm_any_obj_t *obj, int geo)
{
	obj->hdr.calls->pre_edit(obj, geo, 0);
}

RND_INLINE void osm_obj_post_edit_orig(osm_any_obj_t *obj, int tags_changed)
{
	obj->hdr.calls->post_edit(obj, tags_changed);
}

RND_INLINE void osm_obj_draw_attached(osm_any_obj_t *obj, rnd_hid_gc_t gc)
{
	obj->hdr.calls->draw_attached(obj, gc);
}

RND_INLINE void osm_obj_referee_objs(osm_any_obj_t *obj, vtp0_t *dst, int update_att)
{
	obj->hdr.calls->referee_objs(obj, dst, update_att);
}

RND_INLINE osm_any_obj_t *osm_obj_by_id(osm_layer_t *ly, osm_objtype_t obj_type, osm_id_t obj_id)
{
	return htIp_get(&ly->id2obj[obj_type], obj_id);
}

RND_INLINE osm_any_obj_t *osm_obj_dup_into(osm_any_obj_t *obj, osm_layer_t *dst, osm_id_t new_id)
{
	return obj->hdr.calls->dup_into(obj, dst, new_id);
}

RND_INLINE const char *osm_obj_visual_type(osm_any_obj_t *obj)
{
	return obj->hdr.calls->visual_type(obj);
}



#endif
