/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"
#include "tags.h"

#include <string.h>
#include <genht/hash.h>
#include <librnd/core/compat_misc.h>
#include <librnd/core/misc_util.h>

void osm_tags_init(osm_tags_t *tags)
{
	htsp_init(tags, strhash, strkeyeq);
}


void osm_tags_uninit(osm_tags_t *tags)
{
	htsp_entry_t *e;
	for(e = htsp_first(tags); e != NULL; e = htsp_next(tags, e)) {
		free(e->key);
		free(e->value);
	}
	htsp_uninit(tags);
}

void osm_tags_copy(osm_tags_t *dst, const osm_tags_t *src)
{
	htsp_entry_t *e;
	for(e = htsp_first(src); e != NULL; e = htsp_next(src, e))
		htsp_set(dst, rnd_strdup(e->key), rnd_strdup(e->value));
}

void osm_tags_dup(osm_tags_t *dst, const osm_tags_t *src)
{
	osm_tags_init(dst);
	osm_tags_copy(dst, src);
}


void osm_tag_set(osm_tags_t *dst, const char *key, const char *val)
{
	htsp_entry_t *e = htsp_getentry(dst, key);
	if (e != NULL) {
		free(e->value);
		e->value = rnd_strdup(val);
	}
	else
		htsp_set(dst, rnd_strdup(key), rnd_strdup(val));
}

int osm_tags_eq(const osm_tags_t *a, const osm_tags_t *b)
{
	htsp_entry_t *e;

	if (a->used != b->used)
		return 0;

	for(e = htsp_first(a); e != NULL; e = htsp_next(a, e)) {
		const char *vb = htsp_get((osm_tags_t *)b, e->key);
		if (vb == NULL)
			return 0;
		if (strcmp(vb, e->value) != 0)
			return 0;
	}

	return 1;
}


int osm_tags_conflict(const osm_tags_t *a, const osm_tags_t *b)
{
	htsp_entry_t *e;

	if (a->used > b->used)
		rnd_swap(const osm_tags_t *,a, b);

	for(e = htsp_first(a); e != NULL; e = htsp_next(a, e)) {
		const char *vb = htsp_get((osm_tags_t *)b, e->key);
		if (vb == NULL)
			return 0;
		if (strcmp(vb, e->value) != 0)
			return 1;
	}

	return 0;
}


int osm_tags_merge(osm_tags_t *dst, const osm_tags_t *src)
{
	htsp_entry_t *e;

	for(e = htsp_first(src); e != NULL; e = htsp_next(src, e)) {
		const char *vd = htsp_get(dst, e->key);
		if (vd == NULL)
			htsp_set(dst, e->key, rnd_strdup(e->value));
		else {
			/* both src and dst have this key */
			if (strcmp(vd, e->value) != 0)
				return -1; /* collision */
		}
	}
	return 0;
}

