/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"
#include <librnd/config.h>

#include <stdlib.h>
#include <string.h>

#include <librnd/core/actions.h>
#include <librnd/hid/hid_dad.h>
#include <librnd/hid/hid_dad_tree.h>

#include "data.h"
#include "obj_any.h"
#include "search.h"
#include "history.h"

#include "dlg_tags.h"

typedef struct{
	RND_DAD_DECL_NOINIT(dlg)
	unsigned active:1;
	int wobjlst, wtimeline, wdetails;
	gdl_list_t timeline;
} history_ctx_t;

static history_ctx_t history_ctx;

static void history_close_cb(void *caller_data, rnd_hid_attr_ev_t ev)
{
	history_ctx_t *ctx = caller_data;

	RND_DAD_FREE(ctx->dlg);
	osm_timeline_free(&ctx->timeline);
	ctx->active = 0;
}

int hst_cmp(const void *A, const void *B)
{
	osm_any_obj_t * const *a = A, * const *b = B;

	if ((*a)->hdr.type == (*b)->hdr.type)
		return ((*a)->hdr.id < (*b)->hdr.id) ? -1 : +1;

	return ((*a)->hdr.type < (*b)->hdr.type) ? -1 : +1;
}


static void history_osm2dlg_objlst(history_ctx_t *ctx)
{
	rnd_hid_attribute_t *attr;
	rnd_hid_tree_t *tree;
	rnd_hid_row_t *r;
	char *cell[3], *cursor_path = NULL;
	rnd_hid_attr_val_t hv;
	osm_history_t *hst = &osm_hist;
	htIp_entry_t *e;
	vtp0_t all = {0};
	long n;
	int t;

	/* reload the list */
	attr = &ctx->dlg[ctx->wobjlst];
	tree = attr->wdata;

	/* remember cursor */
	r = rnd_dad_tree_get_selected(attr);
	if (r != NULL)
		cursor_path = rnd_strdup(r->cell[0]);

	/* remove existing items */
	rnd_dad_tree_clear(tree);


	/* sort items */
	for(t = 0; t < OSM_OBJ_vmax; t++) {
		for(e = htIp_first(&hst->id2obj[t]); e != NULL; e = htIp_next(&hst->id2obj[t], e)) {
			vtp0_t *lst = e->value;
			if (lst->used > 0)
				vtp0_append(&all, lst->array[lst->used-1]);
		}
	}

	qsort(all.array, all.used, sizeof(osm_any_obj_t *), hst_cmp);

	/* fill in the table */
	cell[1] = NULL;
	for(n = 0; n < all.used; n++) {
		osm_any_obj_t *o = all.array[n];
		cell[0] = rnd_strdup_printf("%s #" OSM_PID, osm_obj_name[o->hdr.type], o->hdr.id);
		r = rnd_dad_tree_append(attr, NULL, cell);
		r->user_data = o;
	}
	vtp0_uninit(&all);

	/* restore cursor */
	if (cursor_path != NULL) {
		hv.str = cursor_path;
		rnd_gui->attr_dlg_set_value(ctx->dlg_hid_ctx, ctx->wobjlst, &hv);
	}

	free(cursor_path);
}

static void hist_append_tags(gds_t *dst, htsp_t *tags, const char *prefix)
{
	htsp_entry_t *e;

	if (tags == NULL)
		return;

	for(e = htsp_first(tags); e != NULL; e = htsp_next(tags, e)) {
		gds_append_str(dst, prefix);
		gds_append_str(dst, e->key);
		gds_append(dst, '=');
		gds_append_str(dst, e->value);
		gds_append(dst, '\n');
	}
}

static void set_details_for_changeset(history_ctx_t *ctx, osm_changeset_t *changeset, vtp0_t *chobjs)
{
	rnd_hid_attr_val_t hv;
	gds_t tmp;
	long n;

	gds_init(&tmp);

	rnd_append_printf(&tmp, "In changeset #" OSM_PID ":\n", changeset->hdr.id);

	gds_append_str(&tmp, "\nTags:\n");
	hist_append_tags(&tmp, &changeset->curr->tags, " ");

	if (chobjs != NULL) {

		gds_append_str(&tmp, "\n\nObjects:\n");
		for(n = 0; n < chobjs->used; n++) {
			osm_any_obj_t *o = chobjs->array[n];
			rnd_append_printf(&tmp, "\n %s #" OSM_PID " v%ld:\n", osm_obj_name[o->hdr.type], o->hdr.id, o->hdr.version);
			hist_append_tags(&tmp, o->hdr.calls->get_tags(o), "  ");
		}
	}

	hv.str = tmp.array;
	rnd_gui->attr_dlg_set_value(ctx->dlg_hid_ctx, ctx->wdetails, &hv);
	gds_uninit(&tmp);
}

static void set_details_for_nothing(history_ctx_t *ctx)
{
	rnd_hid_attr_val_t hv;
	hv.str = "<no changeset selected>";
	rnd_gui->attr_dlg_set_value(ctx->dlg_hid_ctx, ctx->wdetails, &hv);
}

static void history_osm2dlg_details(history_ctx_t *ctx)
{
	rnd_hid_attribute_t *attr = &ctx->dlg[ctx->wobjlst];
	rnd_hid_tree_t *tree = attr->wdata;
	rnd_hid_row_t *r = rnd_dad_tree_get_selected(attr);
	osm_history_t *hst = &osm_hist;
	osm_any_obj_t *obj;
	osm_timeline_t *t;
	char *cell[2];

	osm_timeline_free(&ctx->timeline);

	attr = &ctx->dlg[ctx->wtimeline];
	tree = attr->wdata;

	/* remove existing items */
	rnd_dad_tree_clear(tree);


	if (r == NULL)
		return;

	obj = r->user_data;
	osm_history_insert_timeline_for_obj(hst, &ctx->timeline, obj);

	/* fill in the table */
	cell[1] = NULL;
	for(t = gdl_first(&ctx->timeline); t != NULL; t = t->link.next) {
		cell[0] = rnd_strdup_printf("#" OSM_PID, t->changeset->hdr.id);
		r = rnd_dad_tree_append(attr, NULL, cell);
		r->user_data = t;
	}

	if ((obj != NULL) && (obj->hdr.type == OSM_OBJ_CHANGESET)) {
		vtp0_t *chobjs = NULL;
		set_details_for_changeset(ctx, (osm_changeset_t *)obj, chobjs);
	}
	else
		set_details_for_nothing(ctx);
}

static void history_osm2dlg(history_ctx_t *ctx)
{
	history_osm2dlg_objlst(ctx);
	history_osm2dlg_details(ctx);
}

static void hist_obj_select(rnd_hid_attribute_t *attrib, void *hid_ctx, rnd_hid_row_t *row)
{
	rnd_hid_tree_t *tree = attrib->wdata;
	history_ctx_t *ctx = tree->user_ctx;

	history_osm2dlg_details(ctx);
}


static void hist_timeline_select(rnd_hid_attribute_t *attrib, void *hid_ctx, rnd_hid_row_t *row)
{
	osm_design_t *osm = (osm_design_t *)rnd_multi_get_current();
	osm_history_t *hst = &osm_hist;
	rnd_hid_tree_t *tree = attrib->wdata;
	history_ctx_t *ctx = tree->user_ctx;
	rnd_hid_row_t *r = rnd_dad_tree_get_selected(attrib);
	osm_timeline_t *t;

	osm_history_draw_reset(osm);

	if (r == NULL) {
		set_details_for_nothing(ctx);
		return;
	}

	t = r->user_data;
	set_details_for_changeset(ctx, t->changeset, &t->chobjs);

	osm_history_draw_objs(osm, hst, &t->chobjs);


	/* draw selected object's latest version at t->changeset */
	r = rnd_dad_tree_get_selected(&ctx->dlg[ctx->wobjlst]);
	if (r != NULL) {
		osm_any_obj_t *o, *best = NULL, *obj = r->user_data;
		long n;
		vtp0_t *lst = htIp_get(&hst->id2obj[obj->hdr.type], obj->hdr.id);


		for(n = 0; n < lst->used; n++) {
			o = lst->array[n];
			if (o->hdr.changeset <= t->changeset->hdr.id)
				best = o;
		}

		if (best != NULL)
			osm_history_draw_obj(osm, hst, best);
	}

	rnd_gui->invalidate_all(rnd_gui);
}

static void dlg_hist_jump_to(history_ctx_t *ctx, osm_objtype_t type, osm_id_t id)
{
	rnd_hid_attr_val_t hv;
	char tmp[128];

	rnd_snprintf(tmp, sizeof(tmp), "%s #" OSM_PID, osm_obj_name[type], id);

	hv.str = tmp;
	rnd_gui->attr_dlg_set_value(ctx->dlg_hid_ctx, ctx->wobjlst, &hv);
	history_osm2dlg_details(ctx);
}

static void osm_rnd_dlg_hist(osm_objtype_t type, osm_id_t id)
{
	history_ctx_t *ctx = &history_ctx;
	const char *lhdr[] = {"Object"};
	const char *dhdr[] = {"Changeset"};


	if (ctx->active) {
		TODO("find the entry and put the cursor on it");
		dlg_hist_jump_to(ctx, type, id);
		return;
	}

	RND_DAD_BEGIN_VBOX(ctx->dlg);
		RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);

		RND_DAD_BEGIN_HPANE(ctx->dlg, "left-right");
			RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);

			RND_DAD_BEGIN_VBOX(ctx->dlg); /* left */
				RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);

				RND_DAD_TREE(ctx->dlg, 1, 0, lhdr);
					RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL | RND_HATF_SCROLL);
					ctx->wobjlst = RND_DAD_CURRENT(ctx->dlg);
					RND_DAD_TREE_SET_CB(ctx->dlg, selected_cb, hist_obj_select);
					RND_DAD_TREE_SET_CB(ctx->dlg, ctx, ctx);

			RND_DAD_END(ctx->dlg);

			RND_DAD_BEGIN_HPANE(ctx->dlg, "details-timeline"); /* right */
				RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);

				RND_DAD_BEGIN_VBOX(ctx->dlg);
					RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL | RND_HATF_SCROLL);
					RND_DAD_LABEL(ctx->dlg, "<no changeset selected>");
						ctx->wdetails = RND_DAD_CURRENT(ctx->dlg);
				RND_DAD_END(ctx->dlg);

				RND_DAD_TREE(ctx->dlg, 1, 0, dhdr);
					RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_SCROLL);
					ctx->wtimeline = RND_DAD_CURRENT(ctx->dlg);
					RND_DAD_TREE_SET_CB(ctx->dlg, selected_cb, hist_timeline_select);
					RND_DAD_TREE_SET_CB(ctx->dlg, ctx, ctx);

			RND_DAD_END(ctx->dlg);
		RND_DAD_END(ctx->dlg);
	RND_DAD_END(ctx->dlg);

	ctx->active = 1;

	RND_DAD_DEFSIZE(ctx->dlg, 500, 400);
	RND_DAD_NEW("History", ctx->dlg, "History", ctx, 0, history_close_cb);
	history_osm2dlg(ctx);
	dlg_hist_jump_to(ctx, type, id);
}

const char osm_acts_HistoryDialog[] = "HistoryDialog(object)\n";
const char osm_acth_HistoryDialog[] = "Present the history dialog navigated to an object";
fgw_error_t osm_act_HistoryDialog(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	osm_design_t *osm = (osm_design_t *)RND_ACT_DESIGN;
	osm_layer_t *ly = osm_layer_get_curr(osm);
	osm_any_obj_t *obj;

	if (!osm_hist_inited) {
		rnd_message(RND_MSG_ERROR, "History data not available (empty database)\n");
		return -1;
	}

	if (ly == NULL) {
		rnd_message(RND_MSG_ERROR, "No active layer\n");
		return -1;
	}

	obj = osm_search_obj_xy(ly, osm->crosshair_x, osm->crosshair_y);
	if (ly == NULL) {
		rnd_message(RND_MSG_ERROR, "No object under cursor\n");
		return -1;
	}

	osm_rnd_dlg_hist(obj->hdr.type, obj->hdr.id);
	RND_ACT_IRES(0);
	return 0;
}
