/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"
#include <librnd/config.h>

#include <stdlib.h>
#include <string.h>

#include <librnd/core/actions.h>
#include <librnd/hid/hid_dad.h>
#include <librnd/hid/hid_dad_tree.h>

#include "data.h"
#include "obj_any.h"
#include "select.h"
#include "search.h"

#include "dlg_objlist.h"

struct objlist_ctx_s {
	RND_DAD_DECL_NOINIT(dlg)
	int wlst;

	osm_layer_t *ly;
	vtL0_t *lst;
	void *user_ctx;
};

static void objlist_close_cb(void *caller_data, rnd_hid_attr_ev_t ev)
{
	objlist_ctx_t *ctx = caller_data;

	TODO("free the list");

	RND_DAD_FREE(ctx->dlg);
	free(ctx);
}

static void objlist_osm2dlg(objlist_ctx_t *ctx)
{
	rnd_hid_attribute_t *attr;
	rnd_hid_tree_t *tree;
	rnd_hid_row_t *r, *row;
	char *cell[4], *cursor_path = NULL;
	rnd_hid_attr_val_t hv;
	long n;

	/* reload the list */
	attr = &ctx->dlg[ctx->wlst];
	tree = attr->wdata;

	/* remember cursor */
	r = rnd_dad_tree_get_selected(attr);
	if (r != NULL)
		cursor_path = rnd_strdup(r->cell[0]);

	/* remove existing items */
	rnd_dad_tree_clear(tree);

	cell[3] = NULL;
	for(n = 0; n < ctx->lst->used; n++) {
		osm_objlist_item_t *item = &ctx->lst->array[n];
		cell[0] = rnd_strdup_printf("%ld", n);
		cell[1] = rnd_strdup_printf("%s #" OSM_PID, osm_obj_name[item->type], item->id);
		cell[2] = item->comment == NULL ? NULL : rnd_strdup(item->comment);
		row = rnd_dad_tree_append(attr, NULL, cell);
		row->user_data2.lng = n;
		row->user_data = item;
	}


	/* restore cursor */
	if (cursor_path != NULL) {
		hv.str = cursor_path;
		rnd_gui->attr_dlg_set_value(ctx->dlg_hid_ctx, ctx->wlst, &hv);
	}

	free(cursor_path);
}

void osm_rnd_objlist_osm2dlg(objlist_ctx_t *ctx)
{
	objlist_osm2dlg(ctx);
}

/*
static void referee_select(rnd_hid_attribute_t *attrib, void *hid_ctx, rnd_hid_row_t *row)
{
	rnd_hid_tree_t *tree = attrib->wdata;
	objlist_ctx_t *ctx = tree->user_ctx;
	const char *key = NULL, *val = NULL;
	rnd_hid_attr_val_t hv;

	if (row != NULL) {
		key = row->cell[0];
		val = row->cell[1];
	}

	TODO("set suggestions");

	hv.str = val;
	rnd_gui->attr_dlg_set_value(ctx->dlg_hid_ctx, ctx->wval, &hv);
}
*/

#define GET_ROW \
	rnd_hid_attribute_t *lst_attr = &ctx->dlg[ctx->wlst]; \
	rnd_hid_row_t *row; \
 \
	row = rnd_dad_tree_get_selected(lst_attr); \
	if (row == NULL) { \
		rnd_message(RND_MSG_ERROR, "Need to select a referee first\n"); \
		return; \
	}

static void cb_zoomto(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr)
{
	objlist_ctx_t *ctx = caller_data;
	osm_design_t *osm = ctx->ly->parent;
	fgw_arg_t args[5] = {0}, res;
	GET_ROW;

	args[1].type = FGW_STR; args[1].val.cstr = "obj";
	args[2].type = FGW_STR; args[2].val.cstr = row->cell[1];

	rnd_actionv_bin(&osm->hidlib, "zoom", &res, 3, args);
}

static void cb_select(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr)
{
	objlist_ctx_t *ctx = caller_data;
	osm_any_obj_t *obj;
	osm_objtype_t type;
	osm_id_t id;
	GET_ROW;

	type = osm_str_to_obj_type_id(row->cell[1], NULL, &id);
	if (type == OSM_OBJ_invalid) {
		rnd_message(RND_MSG_ERROR, "Internal error: invalid type:#id pair\n");
		return;
	}
	
	obj = osm_obj_by_id(ctx->ly, type, id);
	if (obj != NULL) {
		osm_obj_select(obj);
		rnd_gui->invalidate_all(rnd_gui);
	}
}

static void cb_unselect(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr)
{
	objlist_ctx_t *ctx = caller_data;
	osm_any_obj_t *obj;
	osm_objtype_t type;
	osm_id_t id;
	GET_ROW;

	type = osm_str_to_obj_type_id(row->cell[1], NULL, &id);
	if (type == OSM_OBJ_invalid) {
		rnd_message(RND_MSG_ERROR, "Internal error: invalid type:#id pair\n");
		return;
	}
	
	obj = osm_obj_by_id(ctx->ly, type, id);
	if (obj != NULL) {
		osm_obj_unselect(obj);
		rnd_gui->invalidate_all(rnd_gui);
	}
}


static void cb_btn(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr)
{
	objlist_ctx_t *ctx = caller_data;
	osm_objlist_item_t *item;
	GET_ROW;

	item = row->user_data;
	if (item->btn_cb != NULL)
		item->btn_cb(ctx->user_ctx, item, attr->val.lng - 1);
}

objlist_ctx_t *osm_rnd_dlg_objlist(osm_layer_t *ly, vtL0_t *lst, void *user_ctx, const char *wintitle, const char **btn_label)
{
	rnd_hid_dad_buttons_t clbtn[] = {{"Close", 0}, {NULL, 0}};
	const char *hdr[] = {"idx", "object", "comment"};
	objlist_ctx_t *ctx;
	int modal = 0;

	if (lst == NULL)
		return NULL;

	ctx = calloc(sizeof(objlist_ctx_t), 1);
	ctx->lst = lst;
	ctx->user_ctx = user_ctx;
	ctx->ly = ly;

	RND_DAD_BEGIN_VBOX(ctx->dlg);
		RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);

			RND_DAD_BEGIN_VBOX(ctx->dlg); /* First tab: referees/coords */
				RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);

				RND_DAD_TREE(ctx->dlg, 3, 0, hdr);
					RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL | RND_HATF_SCROLL);
					ctx->wlst = RND_DAD_CURRENT(ctx->dlg);
				RND_DAD_BEGIN_HBOX(ctx->dlg);
					RND_DAD_LABEL(ctx->dlg, "Referee:");
					RND_DAD_BUTTON(ctx->dlg, "zoom to");
						RND_DAD_CHANGE_CB(ctx->dlg, cb_zoomto);
					RND_DAD_BUTTON(ctx->dlg, "select");
						RND_DAD_CHANGE_CB(ctx->dlg, cb_select);
					RND_DAD_BUTTON(ctx->dlg, "unselect");
						RND_DAD_CHANGE_CB(ctx->dlg, cb_unselect);
					if (btn_label != NULL) {
						const char **s;
						int n, wi;

						for(s = btn_label, n = 1; *s != NULL; s++,n++) {
							RND_DAD_BUTTON(ctx->dlg, *s);
								RND_DAD_CHANGE_CB(ctx->dlg, cb_btn);
								wi = RND_DAD_CURRENT(ctx->dlg);
								ctx->dlg[wi].val.lng = n;
						}
					}
				RND_DAD_END(ctx->dlg);

			RND_DAD_END(ctx->dlg);

		RND_DAD_BUTTON_CLOSES(ctx->dlg, clbtn);
	RND_DAD_END(ctx->dlg);


	RND_DAD_NEW("objlist", ctx->dlg, wintitle, ctx, modal, objlist_close_cb);
	objlist_osm2dlg(ctx);

	return ctx;
}

void osm_objlist_drc_ev(rnd_design_t *hidlib, void *user_data, int argc, rnd_event_arg_t argv[])
{
	static const char *buttons[] = {"Fix it", NULL};
	if (argc > 2)
		osm_rnd_dlg_objlist(argv[1].d.p, argv[2].d.p, NULL, "DRC errors", buttons);
}

static void list_here_cb(void *user_ctx, osm_objlist_item_t *item, int btn_idx)
{
	osm_layer_t *ly = user_ctx;
	osm_design_t *osm = ly->parent;
	osm_any_obj_t *obj = osm_obj_by_id(ly, item->type, item->id);
	fgw_arg_t args[4], res;

	if (obj == NULL) {
		rnd_message(RND_MSG_ERROR, "Object not found on layer\n");
		return;
	}

	switch(btn_idx) {
		case 0: /* props */
			args[1].type = FGW_STR; args[1].val.cstr = "BinObject";
			args[2].type = FGW_PTR; args[2].val.ptr_void = obj;
			rnd_actionv_bin(&osm->hidlib, "propsdialog", &res, 3, args);
			break;
		case 1: /* tags */
			args[1].type = FGW_STR; args[1].val.cstr = "BinObject";
			args[2].type = FGW_PTR; args[2].val.ptr_void = obj;
			rnd_actionv_bin(&osm->hidlib, "tagsdialog", &res, 3, args);
			break;
		default:
			rnd_message(RND_MSG_ERROR, "list_here_cb() invalid button idx: %d\n", btn_idx);
	}
}


const char osm_acts_ListHereDialog[] = "ListHereDialog()";
const char osm_acth_ListHereDialog[] = "Present an list of objects that are near the current location";
fgw_error_t osm_act_ListHereDialog(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	osm_design_t *osm = (osm_design_t *)RND_ACT_DESIGN;
	osm_layer_t *ly = osm_layer_get_curr(osm);
	rnd_coord_t x, y, width, widths[] = {1, 5, 10, 30, 0};
	int n;
	vtp0_t ptrs = {0};
	vtL0_t lst = {0};
	osm_objlist_item_t *i;
	const char *btns[] = {"props", "tags", NULL};

	if (rnd_hid_get_coords("ListHereDialog: click on a location", &x, &y, 0) < 0)
		return 0;

	for(n = 0; (width = widths[n]) != 0; n++) {
		osm_rtree_box_t box;

		width *= osm_slop;
		box.x1 = x - width; box.y1 = y - width;
		box.x2 = x + width; box.y2 = y + width;

		osm_search_append_objs_box(&ptrs, ly, &box);

		if (ptrs.used > 0)
			break;
	}

	if (ptrs.used == 0)
		return 0;

	vtL0_enlarge(&lst, ptrs.used-1);
	for(n = 0, i = lst.array; n < ptrs.used; n++,i++) {
		osm_any_obj_t *obj = ptrs.array[n];
		const char *vty = osm_obj_visual_type(obj);

		i->type = obj->hdr.type;
		i->id = obj->hdr.id;
		i->comment = ((vty == NULL) ? NULL : rnd_strdup(vty));
		i->btn_cb = list_here_cb;
	}
	vtp0_uninit(&ptrs);

	osm_rnd_dlg_objlist(ly, &lst, ly, "ListHere objects", btns);

	/* lst ownership is passed on to the dialog box, don't free it */

	return 0;
}

