/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - server communication
 *  Copyright (C) 2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */


#include "config.h"

#include <libxml/tree.h>
#include <libxml/parser.h>

#include "obj_any.h"
#include "htdr.h"


static int osm_diffresult_apply(osm_layer_t *ly, const char *fn)
{
	xmlDoc *doc;
	xmlNode *root, *n;
	int res = 0;
	osm_rtree_box_t bbox;
	htdr_t hash;

	doc = xmlReadFile(fn, NULL, 0);
	if (doc == NULL) {
		rnd_message(RND_MSG_ERROR, "diff result: xml parsing error on file %s\n", fn);
		return -1;
	}

	root = xmlDocGetRootElement(doc);
	if (xmlStrcmp(root->name, (xmlChar *)"diffResult") != 0) {
		rnd_message(RND_MSG_ERROR, "xml error: root is not <diffResult>\n");
		xmlFreeDoc(doc);
		return -1;
	}

	bbox.x2 = bbox.y2 = RND_COORD_MAX;
	bbox.x1 = bbox.y1 = -RND_COORD_MAX;

	htdr_init(&hash, htdr_keyhash, htdr_keyeq);

	for(n = root->children; n != NULL; n = n->next) {
		osm_objtype_t ot = OSM_OBJ_invalid, i;
		osm_id_t old_id, new_id = 0, new_version = 0;
		htdr_key_t key;
		htdr_value_t val;
		char *s, *end;
		osm_any_obj_t *obj;

		if (strcmp((char *)n->name, "text") == 0)
			continue;

		for(i = 0; i < OSM_OBJ_max; i++) {
			if (strcmp((char *)n->name, osm_obj_name[i]) == 0) {
				ot = i;
				break;
			}
		}

		if (ot == OSM_OBJ_invalid) {
			rnd_message(RND_MSG_ERROR, "xml error: diffres node '%s' is not a known object type\n", n->name);
			res = -1;
			continue;
		}

		s = (char *)xmlGetProp(n, (const xmlChar *)"old_id");
		if (s == NULL) {
			rnd_message(RND_MSG_ERROR, "xml error: diffres node '%s' without old_id\n", n->name);
			res = -1;
			continue;
		}
		old_id = osm_str2id(s, &end);
		if (*end != '\0') {
			rnd_message(RND_MSG_ERROR, "xml error: diffres node '%s' with invalid old_id '%s'\n", n->name, s);
			res = -1;
			continue;
		}

		s = (char *)xmlGetProp(n, (const xmlChar *)"new_id");
		if (s != NULL) {
			new_id = osm_str2id(s, &end);
			if (*end != '\0') {
				rnd_message(RND_MSG_ERROR, "xml error: diffres node '%s' with invalid new_id '%s'\n", n->name, s);
				res = -1;
				continue;
			}
		}

		s = (char *)xmlGetProp(n, (const xmlChar *)"new_version");
		if (s != NULL) {
			new_version = osm_str2id(s, &end);
			if (*end != '\0') {
				rnd_message(RND_MSG_ERROR, "xml error: diffres node '%s' with invalid new_version '%s'\n", n->name, s);
				res = -1;
				continue;
			}
		}

		/* if new_id is 0, it's a delete; if old_id is negative, it's a create */
		rnd_trace("diffres: %d " OSM_PID " " OSM_PID " " OSM_PID "\n",
			ot, old_id, new_id, new_version);

		if (new_id != 0) {
			obj = osm_obj_by_id(ly, ot, old_id);
			if (obj == NULL) {
				rnd_message(RND_MSG_ERROR, "xml error: diffres node '%s' with invalid old_id " OSM_PID ": object not found\n", n->name, old_id);
				continue;
			}
		}
		else {
			long i;

			/* delete objects from the remove list */
			for(i = 0; i < ly->remove_list.used; i++) {
				osm_any_obj_t *remo = ly->remove_list.array[i];
				if ((remo != NULL) && (remo->hdr.type == ot) && (remo->hdr.id == old_id)) {
					vtp0_remove(&ly->remove_list, i, 1);
					osm_obj_free(remo);
					break;
				}
			}
			continue; /* deleted objects are not on layers so no need to remove references or do bbox things */
		}

		osm_rtree_box_bump(&bbox, osm_obj_bbox(obj));
		obj->hdr.calls->diffresult(obj, new_id, new_version);

		key.type = ot;
		key.old_id = old_id;
		val.new_id = new_id;
		val.new_version = new_version;
		val.valid = 1;
		htdr_set(&hash, key, val);
	}

	/* Update al referers within the given bbox */
	{
		osm_rtree_it_t it;
		void *o;

		for(o = osm_rtree_first(&it, &ly->objs, &bbox); o != NULL; o = osm_rtree_next(&it)) {
			osm_any_obj_t *obj = o;

			if (obj->hdr.type == OSM_OBJ_NODE) continue; /* otimization: nodes don't refer to other objects */
			obj->hdr.calls->diffresult_referees(obj, &hash, 0, 0);
		}
	}

	/* Sanity checks */
	if (ly->remove_list.used != 0) {
		long n;
		int warned = 0;
		for(n = ly->remove_list.used - 1; n >= 0; n--) {
			osm_any_obj_t *remo = ly->remove_list.array[n];

			if (remo == NULL) continue;

			if ((remo->hdr.id > 0) && !warned) {
				rnd_message(RND_MSG_ERROR, "osm_diffresult_apply(): objects left on the remove list\n");
				warned = 1;
			}
			vtp0_remove(&ly->remove_list, n, 1);
			osm_obj_free(remo);
		}
	}

	htdr_uninit(&hash);
	xmlFreeDoc(doc);
	return res;
}

