/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - standard tools
 *  Copyright (C) 2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"

#include "data.h"
#include "crosshair.h"
#include <librnd/hid/tool.h>

typedef struct osm_tool_way_s {
	osm_way_t *way;
	rnd_coord_t last_x, last_y;
	rnd_coord_t adjx, adjy;
} osm_tool_way_t;

static osm_tool_way_t way_tool;


static void osm_tool_way_uninit(void)
{
	way_tool.way = NULL;
}

void osm_tool_way_press(rnd_design_t *hl)
{
	osm_design_t *osm = (osm_design_t *)hl;
	osm_layer_t *ly = osm_layer_get_curr(osm);
	osm_node_t *nd = NULL;

	if (ly == NULL) {
		rnd_message(RND_MSG_ERROR, "Select a layer first!\n");
		return;
	}

	if (!ly->editable) {
		rnd_message(RND_MSG_ERROR, "Can't draw way: layer is not editable\n");
		return;
	}

	if ((way_tool.way != NULL) && (osm->crosshair_x == way_tool.last_x) && (osm->crosshair_y == way_tool.last_y)) {
		osm_tool_way_uninit();
		return;
	}

	if (ly == NULL)
		return;

	uundo_freeze_serial(&ly->undo);

	/* prefer existing node when snapped */
	nd = osm->crosshair_snap_node;

	/* if no existing node is found, create new */
	if ((nd == NULL) || (nd->hdr.type != OSM_OBJ_NODE))
		nd = osm_tool_node_insert(ly);

	if (nd != NULL) {
		if (way_tool.way == NULL) {
			way_tool.way = (osm_way_t *)osm_obj_create(ly, OSM_OBJ_WAY, 1);
			assert(way_tool.way != NULL);
		}

		osm_way_modify_append(ly, way_tool.way, nd->hdr.id, 1);
	}

	uundo_unfreeze_serial(&ly->undo);
	uundo_inc_serial(&ly->undo);

	way_tool.last_x = osm->crosshair_x;
	way_tool.last_y = osm->crosshair_y;
	way_tool.adjx = osm->crosshair_x;
	way_tool.adjy = osm->crosshair_y;

	/* stop when closed*/
	if ((nd != NULL) && (way_tool.way->curr->nd.used >= 3) && (nd->hdr.id == way_tool.way->curr->nd.array[0]))
		osm_tool_way_uninit();
}

void osm_tool_way_adjust_attached_objects(rnd_design_t *hl)
{
	osm_design_t *osm = (osm_design_t *)hl;
	way_tool.adjx = osm->crosshair_x;
	way_tool.adjy = osm->crosshair_y;
}

void osm_tool_way_draw_attached(rnd_design_t *hl)
{
	if (way_tool.way != NULL) {
		rnd_hid_set_line_cap(osm_crosshair_gc, rnd_cap_round);
		rnd_hid_set_line_width(osm_crosshair_gc, -1);

		rnd_render->draw_line(osm_crosshair_gc, way_tool.last_x, way_tool.last_y, way_tool.adjx, way_tool.adjy);
	}
}

void osm_tool_way_escape(rnd_design_t *hl)
{
	if (way_tool.way != NULL)
		osm_tool_way_uninit();
	else
		rnd_tool_select_by_name(hl, "arrow");
}


/* XPM */
static const char *way_icon[] = {
/* columns rows colors chars-per-pixel */
"21 21 3 1",
"  c #000000",
". c #6EA5D7",
"o c None",
/* pixels */
"ooooooooooooooooooooo",
"ooooooooooooooo     o",
"ooooooooooooooo ... o",
"ooooooooooooooo . . o",
"oooooooooooooo..... o",
"oooooooooooo...     o",
"ooooooooo...ooooooooo",
"o     ...oooooooooooo",
"o .....oooooooooooooo",
"o . . ooooooooooooooo",
"o ... ooooooooooooooo",
"o     ooooooooooooooo",
"ooooooooooooooooooooo",
"o ooooo oo  ooo ooo o",
"o ooooo oo  ooo ooo o",
"o ooooo o oo ooo o oo",
"oo o o oo oo ooo o oo",
"oo o o oo    oooo ooo",
"oo o o o oooo ooo ooo",
"ooo o oo oooo ooo ooo",
"ooo o oo oooo ooo ooo"
};

rnd_tool_t osm_rnd_tool_way = {
	"way", NULL, NULL, 100, way_icon, RND_TOOL_CURSOR_NAMED("up_arrow"), 0,
	NULL,
	osm_tool_way_uninit,
	osm_tool_way_press,
	NULL,
	osm_tool_way_adjust_attached_objects,
	osm_tool_way_draw_attached,
	NULL,
	NULL,
	osm_tool_way_escape
};
